import math

import numpy as np
import pandas as pd
from SapphireQuant.Config.TradingDay import TradingDayHelper


class StrategyToolBox:
    """
    策略工具箱
    """

    @staticmethod
    def calc_total_and_annual_yield_rate(df_asset, capital, trading_day_per_year, trading_day_num):
        """
        计算总收益率和年华收益率
        :param df_asset:
        :param capital:
        :param trading_day_per_year:
        :param trading_day_num:
        :return:
        """
        yield_rate = df_asset['净利润'].sum() / capital
        if yield_rate < 0:
            annual_yield_rate = -(math.pow(1 + abs(yield_rate), trading_day_per_year / trading_day_num) - 1)
        else:
            annual_yield_rate = math.pow(1 + abs(yield_rate), trading_day_per_year / trading_day_num) - 1

        return yield_rate, annual_yield_rate

    @staticmethod
    def calc_profit_per_hand(df_asset, total_trade_lots):
        """
        计算每手盈利
        :param df_asset:
        :param total_trade_lots:
        :return:
        """
        total_profit = df_asset['净利润'].sum()

        if total_trade_lots != 0:
            profit_per_hand = int(total_profit / total_trade_lots)
        else:
            profit_per_hand = 0
        return profit_per_hand

    @staticmethod
    def calc_max_retrace_rate(df_asset, capital):
        """
        计算最大回撤
        :param df_asset:
        :param capital:
        :return:
        """
        cum_trade_df = df_asset[['净利润']].cumsum(axis=0)
        tp = round(cum_trade_df.loc[len(cum_trade_df) - 1, '净利润'], 0)
        max_retracement = 0
        high_level = 0

        if tp >= 0:
            for v in range(len(cum_trade_df)):
                if cum_trade_df.loc[v, '净利润'] > high_level:
                    high_level = cum_trade_df.loc[v, '净利润']
                temp = cum_trade_df.loc[v, '净利润'] - high_level
                if temp < max_retracement:
                    max_retracement = temp
        elif tp < 0:
            for v in range(len(cum_trade_df)):
                if cum_trade_df.loc[v, '净利润'] < high_level:
                    high_level = cum_trade_df.loc[v, '净利润']
                temp = cum_trade_df.loc[v, '净利润'] - high_level
                if temp > max_retracement:
                    max_retracement = temp

        max_retrace_rate = max_retracement / capital
        return max_retrace_rate

    @staticmethod
    def calc_current_max_retrace_rate(df_asset, capital):
        """
        计算当前最大回撤
        :param df_asset:
        :param capital:
        :return:
        """
        cum_trade_df = df_asset[['净利润']].cumsum(axis=0)
        tp = round(cum_trade_df.loc[len(cum_trade_df) - 1, '净利润'], 0)
        current_retracement = 0
        high_level = 0

        if tp >= 0:
            for v in range(len(cum_trade_df)):
                if cum_trade_df.loc[v, '净利润'] > high_level:
                    high_level = cum_trade_df.loc[v, '净利润']
                temp = cum_trade_df.loc[v, '净利润'] - high_level
                if temp < 0:
                    current_retracement = temp
                else:
                    current_retracement = 0
        elif tp < 0:
            for v in range(len(cum_trade_df)):
                if cum_trade_df.loc[v, '净利润'] < high_level:
                    high_level = cum_trade_df.loc[v, '净利润']
                temp = cum_trade_df.loc[v, '净利润'] - high_level
                if temp > 0:
                    current_retracement = temp
                else:
                    current_retracement = 0

        current_retrace_rate = current_retracement / capital
        return current_retrace_rate

    @staticmethod
    def calc_retrace_ratio(max_retrace_rate, annual_yield_rate):
        """
        计算收益回撤比
        :param max_retrace_rate:
        :param annual_yield_rate:
        :return:
        """
        if max_retrace_rate != 0:
            retrace_ratio = round(
                abs(annual_yield_rate / max_retrace_rate), 2)
        else:
            retrace_ratio = np.nan
        return retrace_ratio

    @staticmethod
    def calc_sharpe_ratio(df_asset, capital):
        """
        计算夏普率
        :param df_asset:
        :param capital:
        :return:
        """
        try:
            return_per_day = df_asset['净利润'] / capital
            sharpe_ratio = return_per_day.mean() / return_per_day.std() * np.sqrt(252)
            return sharpe_ratio
        except Exception as e:
            print('计算夏普率出错:{0}'.format(str(e)))
            return 0

    @staticmethod
    def calc_max_continuous_no_innovation_high_days(df_asset):
        """
        计算最大连续不创新高天数
        @param df_asset:
        """
        high = 0
        days = 0
        max_days = 0
        for index, row in df_asset.iterrows():
            value = row['累计盈亏']
            if value >= high:
                high = value
                if days > max_days:
                    max_days = days
                days = 0
            else:
                days += 1

        if days > max_days:
            max_days = days
        return max_days

    @staticmethod
    def calc_current_max_continuous_no_innovation_high_days(df_asset):
        """
        计算当前最大连续不创新高天数
        @param df_asset:
        """
        high = 0
        days = 0
        for index, row in df_asset.iterrows():
            value = row['累计盈亏']
            if value >= high:
                high = value
                days = 0
            else:
                days += 1
        return days

    @staticmethod
    def calc_max_loss_days(df_asset):
        """
        计算最大连续亏损天数
        @param df_asset:
        """
        days = 0
        max_days = 0
        for index, row in df_asset.iterrows():
            value = row['净利润']
            if value >= 0:
                if days > max_days:
                    max_days = days
                days = 0
            else:
                days += 1

        if days > max_days:
            max_days = days
        return max_days

    @staticmethod
    def compare_trade(df_source, df_target, how='outer', begin_date=None, end_date=None, source_suffix=None, target_suffix=None):
        """
        对比成交
        :param how:
        :param target_suffix:
        :param source_suffix:
        :param df_source:必须包含列名：['交易日', '价格', '合约', '多空', '开平', '时间', '量']
        :param df_target:必须包含列名：['交易日', '价格', '合约', '多空', '开平', '时间', '量']
        :param begin_date:
        :param end_date:
        """

        df_source = df_source[['交易日', '价格', '合约', '多空', '开平', '时间', '量']]
        df_source['时分'] = df_source['时间'].apply(lambda x: x.strftime('%H:%M'))
        df_source['时间'] = df_source['时间'].apply(lambda x: x.strftime('%H:%M:%S'))
        df_source.loc[df_source['开平'] == '平今仓', '开平'] = '平仓'
        if begin_date is not None:
            df_source = df_source[df_source['交易日'] >= begin_date].copy()
        if end_date is not None:
            df_source = df_source[df_source['交易日'] <= end_date].copy()
        if source_suffix is not None:
            df_source.rename(columns={'时间': '时间_{0}'.format(source_suffix), '价格': '价格_{0}'.format(source_suffix), '量': '量_{0}'.format(source_suffix)}, inplace=True)
        else:
            df_source.rename(columns={'时间': '时间_1', '价格': '价格_1', '量': '量_1'}, inplace=True)

        df_target['时分'] = df_target['时间'].apply(lambda x: x.strftime('%H:%M'))
        df_target['时间'] = df_target['时间'].apply(lambda x: x.strftime('%H:%M:%S'))
        df_target.loc[df_target['开平'] == '平今仓', '开平'] = '平仓'
        if begin_date is not None:
            df_target = df_target[df_target['交易日'] >= begin_date].copy()
        if end_date is not None:
            df_target = df_target[df_target['交易日'] <= end_date].copy()
        if source_suffix is not None:
            df_target.rename(columns={'时间': '时间_{0}'.format(target_suffix), '价格': '价格_{0}'.format(target_suffix), '量': '量_{0}'.format(target_suffix)}, inplace=True)
        else:
            df_target.rename(columns={'时间': '时间_2', '价格': '价格_2', '量': '量_2'}, inplace=True)
        df_all = df_source.merge(df_target, on=['交易日', '时分', '合约', '多空', '开平'], how=how)
        df_all = df_all.sort_values(by=['交易日', '时分'], axis=0, ascending=True)
        df_all.reset_index(inplace=True)
        return df_all

    @staticmethod
    def combine_asset(lst_df_data):
        """
        组合Asset
        @param lst_df_data:
        """
        df_combine = None
        for df_asset in lst_df_data:
            if df_combine is None:
                df_combine = df_asset
            else:
                df_combine = df_combine.merge(df_asset, on=['日期'], how='outer')
                df_combine['昨日结存'] = df_combine['昨日结存_x'] + df_combine['昨日结存_y']
                df_combine['动态权益'] = df_combine['动态权益_x'] + df_combine['动态权益_y']
                df_combine['平仓盈亏'] = df_combine['平仓盈亏_x'] + df_combine['平仓盈亏_y']
                df_combine['持仓盈亏'] = df_combine['持仓盈亏_x'] + df_combine['持仓盈亏_y']
                df_combine['净利润'] = df_combine['净利润_x'] + df_combine['净利润_y']
                df_combine['可用资金'] = df_combine['可用资金_x'] + df_combine['可用资金_y']
                df_combine['持仓保证金'] = df_combine['持仓保证金_x'] + df_combine['持仓保证金_y']
                df_combine['手续费'] = df_combine['手续费_x'] + df_combine['手续费_y']
                df_combine = df_combine[['日期', '昨日结存', '动态权益', '平仓盈亏', '持仓盈亏', '净利润', '手续费', '可用资金', '持仓保证金']]
        df_combine['累计盈亏'] = df_combine[['净利润']].cumsum(axis=0)
        df_combine['累计收益率'] = df_combine['累计盈亏'] / df_combine['昨日结存'][0]
        df_combine['累计收益率'] = df_combine['累计收益率'].apply(lambda x: format(x, '.2%'))
        return df_combine

    @staticmethod
    def combine_trade(lst_df_data):
        """
        组合Asset
        @param lst_df_data:
        """
        df_combine = None
        for df_trade in lst_df_data:
            if df_combine is None:
                df_combine = df_trade
            else:
                df_combine = df_combine.concat(df_trade)
        df_combine.sort_values(by=['交易日', '时间'], ascending=True, inplace=True)
        return df_combine

    @staticmethod
    def get_strategy_param(df_param):
        """
        获取策略的参数
        @param df_param:
        @return:
        """
        dic_param = {}
        for index, row in df_param.iterrows():
            dic_param[row['参数名']] = row['参数值']
        return dic_param

    @staticmethod
    def get_strategy_summary(df_asset, df_trade, begin_date, end_date):
        """
        获取策略报告
        :param df_asset:
        :param df_trade:
        :param begin_date:
        :param end_date:
        """
        fund_balance = df_asset.iloc['昨日结存'][0]
        df_asset = df_asset[df_asset['交易日'] >= begin_date]
        df_asset = df_asset[df_asset['交易日'] <= end_date]
        df_trade = df_trade[df_trade['交易日'] >= begin_date]
        df_trade = df_trade[df_trade['交易日'] <= end_date]
        trading_day_num = TradingDayHelper.get_trading_day_count(begin_date, end_date)
        total_trade_lots = df_trade['量'].sum() / 2
        volume_per_day = total_trade_lots / len(df_asset)
        yield_rate = StrategyToolBox.calc_total_and_annual_yield_rate(df_asset, fund_balance, 243, trading_day_num)
        total_yield_rate = yield_rate[0]
        annual_yield_rate = yield_rate[1]
        profit_per_hand = StrategyToolBox.calc_profit_per_hand(df_asset, total_trade_lots)
        max_retrace_rate = StrategyToolBox.calc_max_retrace_rate(df_asset, fund_balance)
        retrace_ratio = StrategyToolBox.calc_retrace_ratio(max_retrace_rate, annual_yield_rate)
        sharpe_ratio = StrategyToolBox.calc_sharpe_ratio(df_asset, fund_balance)
        lst_summary = [
            [begin_date, end_date, trading_day_num, total_trade_lots, volume_per_day,
             total_yield_rate, annual_yield_rate, profit_per_hand, max_retrace_rate, retrace_ratio, sharpe_ratio]]
        df_summary = pd.DataFrame(lst_summary, columns=['开始日期', '截止日期', '交易天数', '总手数', '日均手数',
                                                        '总收益', '年化收益', '每手盈利', '最大回撤', '收益回撤比', '夏普率'])
        return df_summary



import datetime
import re

from SapphireQuant.Config.Instrument.FutureManager import FutureManager
from SapphireQuant.Config.TradingFrame.TradingFrameManager import TradingFrameManager
from SapphireQuant.Core.DateTimeSlice import DateTimeSlice
from SapphireQuant.Core.Enum.EnumExchange import EnumExchange
from SapphireQuant.Core.Enum.EnumMarket import EnumMarket
from SapphireQuant.Core.TimeSlice import TimeSlice
from SapphireQuant.Core.YfTimeHelper import YfTimeHelper


class InstrumentManager:
    """
    合约管理
    """
    FuMainId = "9999"
    FuCurMonthId = "9998"
    FuNextMonthId = "9997"
    FuNextQuarterId = "9996"
    FuNNextQuarterId = "9995"
    FuIndexId = "9990"

    def __init__(self):
        self.__all_instruments = {}
        self.__all_exchanges = {}
        self.__all_products = {}
        self.trading_frame_manager = TradingFrameManager()
        self.future_manger = FutureManager()

    def __getitem__(self, item):
        if item in self.__all_instruments.keys():
            return self.__all_instruments[item]
        else:
            instrument_id = '{0}9999'.format(self.get_product_id(item))
            if instrument_id in self.__all_instruments.keys():
                instrument = self.__all_instruments[instrument_id]
                instrument.id = item
                instrument.instrument_id = item
                return instrument
        return None

    def contains(self, instrument_id):
        """
        是否包含某个合约
        :param instrument_id:
        :return:
        """
        return instrument_id in self.__all_instruments.keys()

    @property
    def all_instruments(self):
        """
        所有的合约列表
        :return:
        """
        return self.__all_instruments

    def get_exchange_by_instrument_id(self, instrument_id):
        """
        获取交易市场
        :param instrument_id:
        :return:
        """
        instrument = self.__all_instruments[instrument_id]
        if instrument is not None:
            return self.get_exchange(instrument.exchange_id)

        return None

    def get_exchange(self, exchange):
        """
        获取交易市场
        :param exchange:
        :return:
        """
        if isinstance(exchange, str):
            if exchange in self.__all_exchanges.keys():
                return self.__all_exchanges[exchange]
        if isinstance(exchange, EnumExchange):
            if exchange == EnumExchange.中金所:
                return self.get_exchange("CFFEX")
            elif exchange == EnumExchange.大商所:
                return self.get_exchange("DCE")
            elif exchange == EnumExchange.上期所:
                return self.get_exchange("SHFE")
            elif exchange == EnumExchange.郑商所:
                return self.get_exchange("CZCE")
            elif exchange == EnumExchange.上证所:
                return self.get_exchange("SH")
            elif exchange == EnumExchange.深交所:
                return self.get_exchange("SZ")
            else:
                raise Exception("非法的交易所代码:" + str(exchange))
        return None

    @property
    def all_exchanges(self):
        """
        所有交易市场
        :return:
        """
        return self.__all_exchanges

    def clear(self):
        """
        清空
        """
        self.__all_instruments.clear()
        self.__all_exchanges.clear()

    def __get_trading_time(self, trading_day, instrument_id):
        instrument = self[instrument_id]
        if (instrument is not None) & (self.trading_frame_manager is not None):
            product_id = InstrumentManager.get_product_id(instrument_id)
            return self.trading_frame_manager.get_trading_time_slices(trading_day, instrument.market, instrument.exchange_id, product_id)

        if instrument is None:
            product_id = InstrumentManager.get_product_id(instrument_id)
            if (product_id == "") | (product_id[0] == "."):
                return None

            product = self.get_future_product(product_id)
            if product is None:
                return None

            if (product.all_slice is not None) & (len(product.all_slie) > 0):
                return product.all_slice

            exchange = self.get_exchange(product.exchange_id)

            if exchange is not None:
                return exchange.all_slice

            return None
        else:
            if instrument.market == EnumMarket.期货:
                future = instrument
                if future is None:
                    return None

                product = self.get_future_product(future.product_id)
                if product is None:
                    return None

                if (product.all_slice is not None) & (len(product.all_slie) > 0):
                    return product.all_slice

                exchange = self.get_exchange(product.exchange_id)

                if exchange is not None:
                    return exchange.all_slice

                return None

    def get_trading_time(self, trading_day, instrument_id, *instrument_ids):
        """
        获取交易时间
        :param trading_day:
        :param instrument_id:
        :param instrument_ids:
        :return:
        """
        source = self.__get_trading_time(trading_day, instrument_id)
        if instrument_ids is None:
            return source
        if len(instrument_ids) == 0:
            return source

        for intersect_instrument_id in instrument_ids:
            target = self.__get_trading_time(trading_day, intersect_instrument_id)
            source = InstrumentManager.intersect(source, target)

        return source

    @staticmethod
    def intersect(source, target):
        """
        交易时间求交集
        :param source:
        :param target:
        :return:
        """
        lst_time_slices = []
        lst_source = []
        lst_target = []
        today = datetime.datetime(datetime.date.today().year, datetime.date.today().month, datetime.date.today().day)
        for temp_time_slice in source:
            lst_source.append(YfTimeHelper.get_date_time_by_trading_day(today, temp_time_slice.begin_time))
            lst_source.append(YfTimeHelper.get_date_time_by_trading_day(today, temp_time_slice.end_time))

        for temp_time_slice in target:
            lst_target.append(YfTimeHelper.get_date_time_by_trading_day(today, temp_time_slice.begin_time))
            lst_target.append(YfTimeHelper.get_date_time_by_trading_day(today, temp_time_slice.end_time))

        i = 0
        j = 0
        while (i < len(lst_source) - 1) & (j < len(lst_target) - 1):
            source_begin_time = lst_source[i]
            source_end_time = lst_source[i + 1]
            target_begin_time = lst_target[j]
            target_end_time = lst_target[j + 1]

            if (source_begin_time < target_begin_time) & (source_end_time < target_begin_time):
                i += 2
                continue

            if (target_begin_time < source_begin_time) & (target_end_time < source_begin_time):
                j += 2
                continue

            if source_begin_time >= target_begin_time:
                begin_time = source_begin_time
            else:
                begin_time = target_begin_time

            if source_end_time >= target_end_time:
                end_time = target_end_time
            else:
                end_time = source_end_time

            temp_time_slice = TimeSlice()
            temp_time_slice.begin_time = datetime.timedelta(hours=begin_time.hour, minutes=begin_time.minute, seconds=begin_time.second)
            temp_time_slice.end_time = datetime.timedelta(hours=end_time.hour, minutes=end_time.minute, seconds=end_time.second)
            lst_time_slices.append(temp_time_slice)

            if source_end_time < target_end_time:
                i += 2
                continue
            if source_end_time > target_end_time:
                j += 2
                continue

            i += 2
            j += 2

        return lst_time_slices

    def get_future_product(self, product_id):
        """
        获取期货产品
        :param product_id:
        :return:
        """
        if product_id in self.__all_products.keys():
            return self.__all_products[product_id]

        return None

    def __merge(self, future_manager):
        for future_id in future_manager.all_instruments.keys():
            if future_id not in self.__all_instruments.keys():
                self.__all_instruments[future_id] = future_manager.all_instruments[future_id]

        for exchange_id in future_manager.all_exchanges.keys():
            if exchange_id not in self.__all_exchanges.keys():
                self.__all_exchanges[exchange_id] = future_manager.all_exchanges[exchange_id]

        for product_id in future_manager.all_products.keys():
            if product_id not in self.__all_products.keys():
                self.__all_products[product_id] = future_manager.all_products[product_id]

    def load(self, config_directory, market):
        """
        按市场加载配置
        :param config_directory:
        :param market:
        :return:
        """
        try:
            self.trading_frame_manager.load(config_directory)

            if market == EnumMarket.期货:
                self.future_manger.load(config_directory)
                self.__merge(self.future_manger)
                return True
        except Exception as e:
            self.clear()
            print(str(e))

        return False

    @staticmethod
    def get_market_by_exchange_id(exchange):
        """
        根据市场获取Market
        :param exchange:
        :return:
        """
        if exchange == EnumExchange.大商所:
            return EnumMarket.期货
        if exchange == EnumExchange.上期所:
            return EnumMarket.期货
        if exchange == EnumExchange.郑商所:
            return EnumMarket.期货
        if exchange == EnumExchange.中金所:
            return EnumMarket.期货
        if exchange == EnumExchange.上证所:
            return EnumMarket.股票
        if exchange == EnumExchange.深交所:
            return EnumMarket.股票
        if exchange == EnumExchange.新交所:
            return EnumMarket.外盘

        return EnumMarket.期货

    def is_normal_open_close_time(self, instrument_id):
        """
        判断是否为
        :param instrument_id:
        :return:
        """
        instrument = self[instrument_id]
        if instrument is None:
            return True

        if instrument.market == EnumMarket.期货:
            future = instrument
            if future is None:
                return True

            product = self.get_future_product(future.product_id)
            if product is None:
                return None

            if (product.all_slice is not None) & (len(product.all_slice) > 0):
                return False

        return True

    def get_living_time_slice(self, trading_day, instrument_id):
        """
        获取实盘交易时间区间
        :param trading_day:
        :param instrument_id:
        :return:
        """
        instrument = self[instrument_id]
        if (instrument is not None) & (self.trading_frame_manager is not None):
            product_id = InstrumentManager.get_product_id(instrument_id)
            return self.trading_frame_manager.get_living_time(trading_day, instrument.market, instrument.exchange_id, product_id)

        if instrument is None:
            product_id = InstrumentManager.get_product_id(instrument_id)
            if (product_id == "") | (product_id[0] == "."):
                return None

            product = self.get_future_product(product_id)
            if product is None:
                return None

            if (product.all_slice is not None) & (len(product.all_slie) > 0):
                time_slice = TimeSlice()
                time_slice.begin_time = product.all_slice[0].begin_time
                time_slice.end_time = product.all_slice[-1].end_time
                return time_slice

            exchange = self.get_exchange(product.exchange_id)

            if exchange is not None:
                time_slice = TimeSlice()
                time_slice.begin_time = exchange.all_slice[0].begin_time
                time_slice.end_time = exchange.all_slice[-1].end_time
                return time_slice

            return None
        else:
            if instrument.market == EnumMarket.期货:
                future = instrument
                if future is None:
                    return None

                product = self.get_future_product(future.product_id)
                if product is None:
                    return None

                if (product.all_slice is not None) & (len(product.all_slie) > 0):
                    time_slice = TimeSlice()
                    time_slice.begin_time = product.all_slice[0].begin_time
                    time_slice.end_time = product.all_slice[-1].end_time
                    return time_slice

                exchange = self.get_exchange(product.exchange_id)

                if exchange is not None:
                    time_slice = TimeSlice()
                    time_slice.begin_time = exchange.all_slice[0].begin_time
                    time_slice.end_time = exchange.all_slice[-1].end_time
                    return time_slice

                return None

    @staticmethod
    def get_product_id(instrument_id):
        """
        获取期货产品ID
        :param instrument_id:
        :return:
        """
        return re.sub(r"\d", "", instrument_id, 0)

    def get_exchange_id(self, instrument_id):
        """
        获取ExchangeId
        :param instrument_id:
        :return:
        """
        instrument = self.__all_instruments[instrument_id]
        if instrument is not None:
            return instrument.exchange_id

        print('未找到{0}对应的合约'.format(instrument_id))
        return None

    def get_living_date_time(self, trading_day, instrument_id):
        """
        获取实盘交易时间区间
        :param trading_day:
        :param instrument_id:
        :return:
        """
        time_slice = self.get_living_time_slice(trading_day, instrument_id)
        date_time_slice = DateTimeSlice()
        date_time_slice.begin_time = YfTimeHelper.get_date_time_by_trading_day(trading_day, time_slice.begin_time)
        date_time_slice.end_time = YfTimeHelper.get_date_time_by_trading_day(trading_day, time_slice.end_time)
        return date_time_slice

    def get_product_ids(self, market, exchange_id):
        """
        获取产品ID
        :param market:
        :param exchange_id:
        :return:
        """
        return self.trading_frame_manager.get_product_ids(market, exchange_id)

    def get_all_product_ids(self, market):
        """
        获取产品ID
        :param market:
        :return:
        """
        return self.trading_frame_manager.get_all_product_ids(market)

    def get_main_contracts(self, market, exchange_id):
        """
        获取主力合约列表
        :param market:
        :param exchange_id:
        :return:
        """
        lst_main_contract = []
        lst_product_id = self.trading_frame_manager.get_product_ids(market, exchange_id)
        for product_id in lst_product_id:
            lst_main_contract.append('{0}9999'.format(product_id))
        return lst_main_contract

    def get_listing_date(self, market, instrument_id):
        """
        获取上市日期
        :param market:
        :param instrument_id:
        :return:
        """
        exchange_id = self.get_exchange_id(instrument_id)
        return self.trading_frame_manager.get_listing_date(market, exchange_id, instrument_id)

    def get_active_date(self, market, instrument_id):
        """
        获取活跃日期
        :param market:
        :param instrument_id:
        :return:
        """
        exchange_id = self.get_exchange_id(instrument_id)
        return self.trading_frame_manager.get_active_date(market, exchange_id, instrument_id)

    def update_future_xml(self, lst_futures):
        """
        更新期货配置文件
        :param lst_futures:
        """
        lst_update_futures = []
        lst_product_ids = self.get_all_product_ids(EnumMarket.期货)
        for future in lst_futures:
            if future.product_id in lst_product_ids:
                lst_update_futures.append(future)
        self.future_manger.update(lst_update_futures)

# instrument_manager = InstrumentManager()
#
# config_dir = "D:\WorkSpace\CarlSnow\Python\QiDataProcessing\QiDataProcessing\Profiles"
# instrument_manager.load(config_dir, EnumMarket.期货)
# for instrument_id in instrument_manager.all_instruments.keys():
#     print(instrument_manager.all_instruments[instrument_id].to_string())
#
# for exchange_id in instrument_manager.all_exchanges.keys():
#     print(exchange_id)
#
# instrument_id = "IF9999"
# print("GetItem:" + instrument_manager[instrument_id].to_string())
# print("HasInstrument:" + str(instrument_manager.has_instrument(instrument_id)))
# #
# exchange_id = "CFFEX"
# exchange = instrument_manager.get_exchange(exchange_id)
# print("GetExchangeById:" + str(exchange))
#
# exchange = instrument_manager.get_exchange(EnumExchange.中金所)
# print("GetExchangeByEnum:" + str(exchange))
#
# instrument_id = "ag9999"
# trading_date = datetime.datetime(2019, 10, 10)
# lst_time_slice = instrument_manager.get_trading_time(trading_date, instrument_id)
# print("get_trading_time:")
# for time_slice in lst_time_slice:
#     print(time_slice.to_string())
#
# product = instrument_manager.get_future_product("IF")
# print("get_future_product:" + product.to_string())
#
# time_slice = instrument_manager.get_living_time(trading_date, instrument_id)
# print("get_living_time:" + time_slice.to_string())
#
# product_id = instrument_manager.get_product_id("IF9999")
# print("get_product_id:" + product_id)
#
# date_time_slice = instrument_manager.get_living_date_time(trading_date, instrument_id)
# print("get_living_date_time:" + date_time_slice.begin_time.strftime('%Y/%m/%d %H:%M:%S') + "-" + date_time_slice.end_time.strftime('%Y/%m/%d %H:%M:%S'))



import abc
import datetime
import os
import uuid

from SapphireQuant.DataStorage.Cache.FutureCache import EnumLiquidationType
from SapphireQuant.Lib import FileToolBox


class BaseStrategy:
    """
    策略基类
    """

    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.function = kwargs['function']
        if 'mongodb_client' in kwargs.keys():
            self.mongodb_client = kwargs['mongodb_client']
        else:
            self.mongodb_client = None
        self.record_dir = kwargs['record_dir']
        self.product_code = kwargs['product_code']
        self.strategy_id = kwargs['strategy_id']
        self.drive_code = kwargs['drive_code']
        self.subscribe_futures = kwargs['future_codes']

        self.start_simulate_date = kwargs['start_simulate_date']
        self.end_simulate_date = kwargs['end_simulate_date']
        self.trading_date = self.start_simulate_date
        self.data_handler = kwargs['data_handler']
        self.fund_balance = kwargs['fund_balance']
        self.instrument_manager = self.data_handler.instrument_manager
        self.transaction_rate = kwargs['transaction_rate']
        self.long_margin_rate = kwargs['long_margin_rate']
        self.short_margin_rate = kwargs['short_margin_rate']

        self.liquidation_type = EnumLiquidationType.收盘价
        if 'liquidation_type' in kwargs.keys():
            self.liquidation_type = kwargs['liquidation_type']

        self.future_cache = kwargs['future_cache']

    @property
    def tick_now(self):
        """

        :return:
        """
        return self.function.now

    @property
    def df_asset(self):
        """

        :return:
        """
        return self.function.df_asset

    @property
    def df_order(self):
        """

        :return:
        """
        return self.function.df_order

    @property
    def df_trade(self):
        """

        :return:
        """
        return self.function.df_trade

    @property
    def df_position(self):
        """

        :return:
        """
        return self.function.df_position

    @property
    def df_summary(self):
        """

        :return:
        """
        return self.function.df_summary

    @property
    def all_futures(self):
        """

        :return:
        """
        return self.function.all_futures

    @property
    def last_bar_map(self):
        """

        :return:
        """
        return self.function.last_bar_map

    def add_indicator(self, indicator):
        """
        添加指标
        :param indicator:
        """
        self.function.add_indicator(indicator)

    def send_order(self, instrument_id, direction, open_close, volume, price, order_time=None):
        """
        发送委托
        :param order_time:
        :param instrument_id:
        :param direction:
        :param open_close:
        :param volume:
        :param price:
        """
        if order_time is None:
            if instrument_id in self.last_bar_map.keys():
                last_bar = self.last_bar_map[instrument_id]
                if last_bar is not None:
                    if last_bar.open == price:
                        order_time = last_bar.begin_time
                    elif last_bar.close == price:
                        order_time = last_bar.end_time
                    else:
                        order_time = self.tick_now
                else:
                    order_time = self.tick_now

        order = self.function.send_future_order('', uuid.uuid1(), self.product_code, self.strategy_id,
                                                instrument_id, 'exchange_id', price, volume, direction, open_close, order_time=order_time)
        return order

    @abc.abstractmethod
    def on_start(self):
        """
        回测开始
        """
        pass

    @abc.abstractmethod
    def on_init(self):
        """
        初始化
        """
        pass

    @abc.abstractmethod
    def on_bar(self, bar):
        """
        on_bar触发
        :param bar:
        """
        pass

    @abc.abstractmethod
    def on_tick(self, tick):
        """
        on_tick触发
        :param tick:
        """
        pass

    @abc.abstractmethod
    def on_future_order(self, order):
        """
        委托回报
        :param order:
        """
        pass

    @abc.abstractmethod
    def on_future_order_canceled(self, order):
        """
        撤单成功回报
        :param order:
        """
        pass

    @abc.abstractmethod
    def on_future_cancel_order_failed(self, order_guid, msg):
        """
        撤单失败回报
        :param order_guid:
        :param msg:
        """
        pass

    @abc.abstractmethod
    def on_future_order_rejected(self, order):
        """
        委托拒绝回报
        :param order:
        """
        pass

    @abc.abstractmethod
    def on_future_trade(self, trade):
        """
        成交回报
        :param trade:
        """
        pass

    @abc.abstractmethod
    def on_exit(self):
        """
        当日结束
        """
        pass

    @abc.abstractmethod
    def on_end(self):
        """
        回测结束
        """
        pass

    def get_bar_series_by_date_time(self, market, instrument_id, interval, bar_type, begin_time):
        """
        根据时间获取K线
        :param market:
        :param instrument_id:
        :param interval:
        :param bar_type:
        :param begin_time:
        :return:
        """
        return self.function.get_bar_series_by_date_time(market, instrument_id, interval, bar_type, begin_time)

    def get_bar_series_by_length(self, market: object, instrument_id: object, interval: object, bar_type: object, max_length: object) -> object:
        """
        根据回溯根数获取K线
        :param market:
        :param instrument_id:
        :param interval:
        :param bar_type:
        :param max_length:
        :return:
        """
        return self.function.get_bar_series_by_length(market, instrument_id, interval, bar_type, max_length)

    def save_strategy(self, excel_record_path, fig_record_path):
        """
        保存策略相关回测信息
        """
        try:
            dic_data = {
                'Asset': self.function.df_asset,
                'Order': self.function.df_order,
                'Trade': self.function.df_trade,
                'Position': self.function.df_position,
                'PositionDetail': self.function.df_position_detail,
                'Summary': self.function.df_summary,
                'Param': self.function.df_param,
            }
            FileToolBox.save_excel(excel_record_path, **dic_data)
            self.function.df_asset['累计收益率(%)'] = self.function.df_asset['累计收益率'].str.strip('%').astype(float) / 100
            FileToolBox.save_fig(fig_record_path, self.function.df_asset, '日期', ['累计收益率(%)'])
        except Exception as e:
            print(str(e))

    def save_mongodb(self, db_name, collection_name, account_id):
        """
        保存记录到数据库
        @param db_name:
        @param collection_name:
        @param account_id:
        """
        try:
            record = {
                'version': 1.0,
                'drive_code': self.drive_code,
                'start_simulate_date': self.start_simulate_date,
                'end_simulate_date': self.end_simulate_date,
                'account_id': account_id,
                '_asset': self.function.df_asset.to_dict('records'),
                'trade': self.function.df_trade.to_dict('records'),
                'summary': self.function.df_summary.to_dict('records')[0],
                'param': self.function.df_param.set_index(['参数名']).T.to_dict('records')[0],
                'update_time': datetime.datetime.now(),
                'run_seconds': self.function.run_seconds
            }
            db = self.mongodb_client[db_name]
            collection = db[collection_name]
            collection.insert_one(record)
        except Exception as e:
            print(str(e))
            print('保存数据库出错，自动转存文件格式保存......')
            excel_record_path = os.path.join(self.record_dir, '{0}.xlsx'.format(account_id))
            fig_record_path = os.path.join(self.record_dir, '{0}.png'.format(account_id))
            self.save_strategy(excel_record_path, fig_record_path)

    def get_future_position(self, instrument_id, direction):
        """
        获取期货持仓
        :param instrument_id:
        :param direction:
        :return:
        """
        return self.function.get_future_position('', self.product_code, self.strategy_id, instrument_id, direction)



from datetime import datetime, timedelta

from SapphireQuant.Config.TradingDay import TradingDayHelper

from SapphireQuant.Config.TradingFrame import FutureTradingTime
from SapphireQuant.Core.Enum import EnumMarket, EnumOrderPriceType, EnumOrderStatus, EnumOrderRejectReason, EnumOrderType, EnumHedgeFlag, EnumStrategyWorkMode, EnumBuySell, EnumOpenClose, \
    EnumPositionDirection, EnumPositionType
from SapphireQuant.Core.Futures import FutureOrder
from SapphireQuant.Strategy import StrategyToolBox
import pandas as pd


class Function:
    """
    方法句柄
    """

    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.data_handler = kwargs['data_handler']
        self._trade_handler = kwargs['trade_handler']

        self.instrument_manager = self.data_handler.instrument_manager
        self.fund_balance = kwargs['fund_balance']
        self._future_codes = kwargs['future_codes']
        self._drive_code = kwargs['drive_code']
        self._work_mode = EnumStrategyWorkMode.simulate
        self._start_simulate_date = kwargs['start_simulate_date']
        self._end_simulate_date = kwargs['end_simulate_date']
        self._start_simulate_date = TradingDayHelper.get_first_trading_day(self._start_simulate_date)
        self._end_simulate_date = TradingDayHelper.get_last_trading_day(self._end_simulate_date)

        if 'work_mode' in kwargs.keys():
            self._work_mode = kwargs['work_mode']

        self.all_futures = {}
        self._gen_future_series(self._future_codes)

        self._indicators = []
        self.future_trading_time = None
        self._trading_date = None
        self._now = self._start_simulate_date
        self._last_bar_map = {}

        self.df_asset = None
        self.df_order = None
        self.df_trade = None
        self.df_position = None
        self.df_position_detail = None

        self.df_summary = None
        self.df_param = None
        self.lst_summary = None

        self.trading_day_num = None
        self.total_trade_lots = None
        self.volume_per_day = None
        self.yield_rate = None
        self.total_yield_rate = None
        self.annual_yield_rate = None
        self.profit_per_hand = None
        self.max_retrace_rate = None
        self.retrace_ratio = None
        self.sharpe_ratio = None

    def day_begin(self, trading_date):
        """

        :param trading_date:
        """
        self._trading_date = trading_date
        self.data_handler.trading_day = trading_date
        lst_time_slices = self.get_trading_time(self._drive_code)
        self.future_trading_time = FutureTradingTime(self._drive_code, self._trading_date, lst_time_slices)
        if self.future_trading_time.if_night_flag:
            self._now = self.future_trading_time.night_open + timedelta(minutes=-1)
        else:
            self._now = self.future_trading_time.day_am_open + timedelta(minutes=-1)

    def day_end(self):
        """
        交易日结束
        """

    def add_indicator(self, indicator):
        """

        :param indicator:
        """
        indicator.calculate()
        self._indicators.append(indicator)

    def get_trading_time(self, instrument_id):
        """
        获取品种当日的交易区间
        :param instrument_id:
        :return:
        """
        return self.instrument_manager.get_trading_time(self._trading_date, instrument_id)

    @property
    def last_bar_map(self):
        """

        :return:
        """
        return self._last_bar_map

    @property
    def trading_date(self):
        """
        交易日
        :return:
        """
        return self._trading_date

    @property
    def drive_code(self):
        """
        触发品种
        :return:
        """
        return self._drive_code

    @property
    def start_simulate_date(self):
        """

        :return:
        """
        return self._start_simulate_date

    @property
    def end_simulate_date(self):
        """

        :return:
        """
        return self._end_simulate_date

    @property
    def now(self):
        """
        当前时间
        :return:
        """
        return self._now

    def _gen_future_series(self, future_codes):
        for instrument_id in future_codes:
            if instrument_id not in self.all_futures.keys():
                self.all_futures[instrument_id] = self.instrument_manager[instrument_id]

    def on_tick(self, tick):
        """
        Tick触发
        :param tick:
        """
        self._now = tick.date_time
        self.data_handler.on_tick(tick)
        instrument_id = tick.instrument_id.split('.')[0]
        self.all_futures[instrument_id].last_tick = tick
        if self._work_mode == EnumStrategyWorkMode.simulate:
            if self._drive_code in tick.instrument_id:
                for indicator in self._indicators:
                    indicator.calculate()
        else:
            for indicator in self._indicators:
                indicator.calculate()

    def on_bar(self, bar):
        """
        Bar触发
        :param bar:
        """
        self._now = bar.begin_time
        self._last_bar_map[bar.instrument_id] = bar
        self.data_handler.on_bar(bar)
        if self._work_mode == EnumStrategyWorkMode.simulate:
            if self._drive_code in bar.instrument_id:
                for indicator in self._indicators:
                    indicator.calculate()
        else:
            for indicator in self._indicators:
                indicator.calculate()

    def get_bar_series_by_date_time(self, market, instrument_id, interval, bar_type, begin_time):
        """
        根据时间获取K线
        :param market:
        :param instrument_id:
        :param interval:
        :param bar_type:
        :param begin_time:
        :return:
        """
        return self.data_handler.get_bar_series_by_date_time(market, instrument_id, interval, bar_type, begin_time, self._now)

    def get_bar_series_by_length(self, market, instrument_id, interval, bar_type, max_length) -> object:
        """
        根据回溯根数获取K线
        :param market:
        :param instrument_id:
        :param interval:
        :param bar_type:
        :param max_length:
        :return:
        """
        return self.data_handler.get_bar_series_by_length(market, instrument_id, interval, bar_type, max_length, self._now)

    def send_future_order(self, logic_account_name, task_guid, product_code, strategy_id,
                          instrument_id, exchange_id, price, volume, direction, open_close,
                          price_type=EnumOrderPriceType.限价,
                          # time_force = EnumOrderTimeForce.当日有效,
                          hedge_flag=EnumHedgeFlag.投机,
                          order_time=None):
        """
        发送委托
        :param order_time:
        :param logic_account_name:
        :param task_guid:
        :param product_code:
        :param strategy_id:
        :param instrument_id:
        :param exchange_id:
        :param price:
        :param volume:
        :param direction:
        :param open_close:
        :param price_type:
        :param hedge_flag:
        """
        order = FutureOrder()
        order.order_sys_id = ''
        order.product_code = product_code
        order.strategy_id = strategy_id
        order.task_guid = task_guid
        order.instrument_id = instrument_id
        order.instrument_name = ''
        order.instrument_type = EnumMarket.期货
        order.exchange_id = exchange_id
        order.direction = direction
        order.open_close = open_close
        order.hedge_flag = hedge_flag
        order.order_status = EnumOrderStatus.未知
        order.price_type = price_type
        order.broker_id = ''
        order.investor_id = ''
        order.user_id = ''
        # order.time_force = time_force
        order.order_volume = volume
        order.order_price = price
        if order_time is not None:
            order.order_time = order_time

        order.volume_traded = 0
        order.volume_canceled = 0
        order.cancel_time = None

        order.order_reject_reason = EnumOrderRejectReason.未知
        order.match_price = 0
        order.match_amount = 0
        order.frozen_amount = 0
        order.transaction_cost = 0
        order.order_type = EnumOrderType.normal
        order.local_order_time = None
        order.status_msg = ''
        order.remark = ''
        return self._trade_handler.send_future_order(logic_account_name, order)

    def get_future_position(self, logic_account_name, product_code, strategy_id, instrument_id, direction):
        """
        获取期货持仓
        :param logic_account_name:
        :param product_code:
        :param strategy_id:
        :param instrument_id:
        :param direction:
        """
        self._trade_handler.get_future_position(logic_account_name, product_code, strategy_id, instrument_id, direction)

    def on_end(self):
        """
        生成策略报告
        """
        self.df_asset = self._trade_handler.get_all_future_assets()
        self.df_order = self._trade_handler.get_all_future_orders()
        self.df_trade = self._trade_handler.get_all_future_trades()
        self.df_position = self._trade_handler.get_all_future_positions()
        self.df_position_detail = self._trade_handler.get_all_future_position_details()

        self.trading_day_num = TradingDayHelper.get_trading_day_count(self.start_simulate_date, self.end_simulate_date)
        self.total_trade_lots = self.df_trade['量'].sum() / 2
        self.volume_per_day = self.total_trade_lots / len(self.df_asset)
        self.yield_rate = StrategyToolBox.calc_total_and_annual_yield_rate(self.df_asset, self.fund_balance, 243, self.trading_day_num)
        self.total_yield_rate = self.yield_rate[0]
        self.annual_yield_rate = self.yield_rate[1]
        self.profit_per_hand = StrategyToolBox.calc_profit_per_hand(self.df_asset, self.total_trade_lots)
        self.max_retrace_rate = StrategyToolBox.calc_max_retrace_rate(self.df_asset, self.fund_balance)
        self.retrace_ratio = StrategyToolBox.calc_retrace_ratio(self.max_retrace_rate, self.annual_yield_rate)
        self.sharpe_ratio = StrategyToolBox.calc_sharpe_ratio(self.df_asset, self.fund_balance)
        self.lst_summary = [
            [self.drive_code, self.start_simulate_date, self.end_simulate_date, self.trading_day_num, self.total_trade_lots, self.volume_per_day,
             self.total_yield_rate, self.annual_yield_rate, self.profit_per_hand, self.max_retrace_rate, self.retrace_ratio, self.sharpe_ratio]]
        self.df_summary = pd.DataFrame(self.lst_summary, columns=['品种', '开始日期', '截止日期', '交易天数', '总手数', '日均手数',
                                                                  '总收益', '年化收益', '每手盈利', '最大回撤', '收益回撤比', '夏普率'])

        lst_param = []
        for key in self.kwargs:
            if key not in ['factor_map', 'data_handle', 'mongodb_client']:
                lst_param.append([key, str(self.kwargs[key])])
        self.df_param = pd.DataFrame(lst_param, columns=['参数名', '参数值'])



# -*- coding:utf-8 -*-
"""
Description:
    Binary Writer

Usage:
    from neo.Core.IO.BinaryWriter import BinaryWriter
"""
import sys
import os
import inspect
import struct
import binascii


def swap32(i):
    """
    Change the endianness from little endian to big endian.
    Args:
        i (int):

    Returns:
        int:
    """
    return struct.unpack("<I", struct.pack(">I", i))[0]


def convert_to_u_int160(value):
    """
    Convert an int value to a 10 bytes binary string value.
    Note: the return value is not really 160 bits, nor is it of the neo.Core.UInt160 type

    Args:
        value (int): number to convert.

    Returns:
        str:
    """
    return bin(value + 2 ** 20)[-20:]


def convert_to_u_int256(value):
    """
    Convert an int value to a 16 bytes binary string value.
    Note: the return value is not really 256 bits, nor is it of the neo.Core.UInt256 type

    Args:
        value (int): number to convert.

    Returns:
        str:
    """
    return bin(value + 2 ** 32)[-32:]


class BinaryWriter(object):
    """docstring for BinaryWriter"""

    def __init__(self, stream):
        """
        Create an instance.

        Args:
            stream (BytesIO): a stream to operate on. i.e. a neo.IO.MemoryStream or raw BytesIO.
        """
        super(BinaryWriter, self).__init__()
        self.stream = stream

    def write_byte(self, value):
        """
        Write a single byte to the stream.

        Args:
            value (bytes, str or int): value to write to the stream.
        """
        if type(value) is bytes:
            self.stream.write(value)
        elif type(value) is str:
            self.stream.write(value.encode('utf-8'))
        elif type(value) is int:
            self.stream.write(bytes([value]))
        elif type(value) is bool:
            if value:
                self.stream.write(bytes([1]))
            else:
                self.stream.write(bytes([0]))

    def write_bytes(self, value, unhex=True):
        """
        Write a `bytes` type to the stream.

        Args:
            value (bytes): array of bytes to write to the stream.
            unhex (bool): (Default) True. Set to unhexlify the stream. Use when the bytes are not raw bytes; i.e. b'aabb'

        Returns:
            int: the number of bytes written.
        """
        if unhex:
            try:
                value = binascii.unhexlify(value)
            except binascii.Error:
                pass
        return self.stream.write(value)

    def pack(self, fmt, data):
        """
        Write bytes by packing them according to the provided format `fmt`.
        For more information about the `fmt` format see: https://docs.python.org/3/library/struct.html

        Args:
            fmt (str): format string.
            data (object): the data to write to the raw stream.

        Returns:
            int: the number of bytes written.
        """
        return self.write_bytes(struct.pack(fmt, data), unhex=False)

    def write_char(self, value):
        """
        Write a 1 byte character value to the stream.

        Args:
            value: value to write.

        Returns:
            int: the number of bytes written.
        """
        return self.pack('c', value)

    def write_float(self, value, endian="<"):
        """
        Pack the value as a float and write 4 bytes to the stream.

        Args:
            value (number): the value to write to the stream.
            endian (str): specify the endianness. (Default) Little endian ('<'). Use '>' for big endian.

        Returns:
            int: the number of bytes written.
        """
        return self.pack('%sf' % endian, value)

    def write_double(self, value, endian="<"):
        """
        Pack the value as a double and write 8 bytes to the stream.

        Args:
            value (number): the value to write to the stream.
            endian (str): specify the endianness. (Default) Little endian ('<'). Use '>' for big endian.

        Returns:
            int: the number of bytes written.
        """
        return self.pack('%sd' % endian, value)

    def write_int8(self, value, endian="<"):
        """
        Pack the value as a signed byte and write 1 byte to the stream.

        Args:
            value:
            endian (str): specify the endianness. (Default) Little endian ('<'). Use '>' for big endian.

        Returns:
            int: the number of bytes written.
        """
        return self.pack('%sb' % endian, value)

    def write_u_int8(self, value, endian="<"):
        """
        Pack the value as an unsigned byte and write 1 byte to the stream.

        Args:
            value:
            endian (str): specify the endianness. (Default) Little endian ('<'). Use '>' for big endian.

        Returns:
            int: the number of bytes written.
        """
        return self.pack('%sB' % endian, value)

    def write_bool(self, value):
        """
        Pack the value as a bool and write 1 byte to the stream.

        Args:
            value: the boolean value to write.

        Returns:
            int: the number of bytes written.
        """
        return self.pack('?', value)

    def write_int16(self, value, endian="<"):
        """
        Pack the value as a signed integer and write 2 bytes to the stream.

        Args:
            value:
            endian (str): specify the endianness. (Default) Little endian ('<'). Use '>' for big endian.

        Returns:
            int: the number of bytes written.
        """
        return self.pack('%sh' % endian, value)

    def write_u_int16(self, value, endian="<"):
        """
        Pack the value as an unsigned integer and write 2 bytes to the stream.

        Args:
            value:
            endian (str): specify the endianness. (Default) Little endian ('<'). Use '>' for big endian.

        Returns:
            int: the number of bytes written.
        """
        return self.pack('%sH' % endian, value)

    def write_int32(self, value, endian="<"):
        """
        Pack the value as a signed integer and write 4 bytes to the stream.

        Args:
            value:
            endian (str): specify the endianness. (Default) Little endian ('<'). Use '>' for big endian.

        Returns:
            int: the number of bytes written.
        """
        return self.pack('%si' % endian, value)

    def write_u_int32(self, value, endian="<"):
        """
        Pack the value as an unsigned integer and write 4 bytes to the stream.

        Args:
            value:
            endian (str): specify the endianness. (Default) Little endian ('<'). Use '>' for big endian.

        Returns:
            int: the number of bytes written.
        """
        return self.pack('%sI' % endian, value)

    def write_int64(self, value, endian="<"):
        """
        Pack the value as a signed integer and write 8 bytes to the stream.

        Args:
            value:
            endian (str): specify the endianness. (Default) Little endian ('<'). Use '>' for big endian.

        Returns:
            int: the number of bytes written.
        """
        return self.pack('%sq' % endian, value)

    def write_u_int64(self, value, endian="<"):
        """
        Pack the value as an unsigned integer and write 8 bytes to the stream.

        Args:
            value:
            endian (str): specify the endianness. (Default) Little endian ('<'). Use '>' for big endian.

        Returns:
            int: the number of bytes written.
        """
        return self.pack('%sQ' % endian, value)

    # def write_u_int160(self, value):
    #     """
    #     Write a UInt160 type to the stream.
    #
    #     Args:
    #         value (UInt160):
    #
    #     Raises:
    #         TypeError: when `value` is not of neo.Core.UInt160 type.
    #     """
    #     if type(value) is UInt160:
    #         value.Serialize(self)
    #     else:
    #         raise TypeError("Value must be UInt160 instance.")
    #
    # def WriteUInt256(self, value):
    #     """
    #     Write a UInt256 type to the stream.
    #
    #     Args:
    #         value (UInt256):
    #
    #     Raises:
    #         TypeError: when `value` is not of neo.Core.UInt256 type.
    #     """
    #     if type(value) is UInt256:
    #         value.Serialize(self)
    #     else:
    #         raise TypeError("Value must be UInt256 instance.")

    def write_var_int(self, value, endian="<"):
        """
        Write an integer value in a space saving way to the stream.
        Read more about variable size encoding here: http://docs.neo.org/en-us/node/network-protocol.html#convention

        Args:
            value (int):
            endian (str): specify the endianness. (Default) Little endian ('<'). Use '>' for big endian.

        Returns:
            int: the number of bytes written.

        Raises:
            TypeError: if `value` is not of type int.
            ValueError: if `value` is < 0.
        """
        if not isinstance(value, int):
            raise TypeError(f'{value} not int type.')

        if value < 0:
            raise ValueError(f'{value} too small.')

        elif value < 0xfd:
            return self.write_byte(value)

        elif value <= 0xffff:
            self.write_byte(0xfd)
            return self.write_u_int16(value, endian)

        elif value <= 0xFFFFFFFF:
            self.write_byte(0xfe)
            return self.write_u_int32(value, endian)

        else:
            self.write_byte(0xff)
            return self.write_u_int64(value, endian)

    # def write_var_bytes(self, value, endian="<"):
    #     """
    #     Write an integer value in a space saving way to the stream.
    #     Read more about variable size encoding here: http://docs.neo.org/en-us/node/network-protocol.html#convention
    #
    #     Args:
    #         value (bytes):
    #         endian (str): specify the endianness. (Default) Little endian ('<'). Use '>' for big endian.
    #
    #     Returns:
    #         int: the number of bytes written.
    #     """
    #     length = len(value)
    #     self.write_var_int(length, endian)
    #
    #     return self.write_bytes(value, unhex=False)

    def write_var_string(self, value, encoding="utf-8"):
        """
        Write a string value to the stream.
        Read more about variable size encoding here: http://docs.neo.org/en-us/node/network-protocol.html#convention

        Args:
            value (string): value to write to the stream.
            encoding (str): string encoding format.
        """
        if type(value) is str:
            value = value.encode(encoding)

        length = len(value)
        ba = bytearray(value)
        byts = binascii.hexlify(ba)
        string = byts.decode(encoding)
        self.write_var_int(length)
        self.write_bytes(string)

    def write_fixed_string(self, value, length):
        """
        Write a string value to the stream.

        Args:
            value (str): value to write to the stream.
            length (int): length of the string to write.

        Raises:
            ValueError: if the input `value` length is longer than the fixed `length`
        """
        towrite = value.encode('utf-8')
        slen = len(towrite)
        if slen > length:
            raise ValueError("String '{value}' length is longer than fixed length: {length}")
        self.write_bytes(towrite)
        diff = length - slen

        while diff > 0:
            self.write_byte(0)
            diff -= 1

    # def write_serializable_array(self, array):
    #     """
    #     Write an array of serializable objects to the stream.
    #
    #     Args:
    #         array(list): a list of serializable objects. i.e. extending neo.IO.Mixins.SerializableMixin
    #     """
    #     if array is None:
    #         self.write_byte(0)
    #     else:
    #         self.write_var_int(len(array))
    #         for item in array:
    #             item.Serialize(self)

    def write2000256list(self, arr):
        """
        Write an array of 64 byte items to the stream.

        Args:
            arr (list): a list of 2000 items of 64 bytes in size.
        """
        for item in arr:
            ba = bytearray(binascii.unhexlify(item))
            ba.reverse()
            self.write_bytes(ba)

    # def write_hashes(self, arr):
    #     """
    #     Write an array of hashes to the stream.
    #
    #     Args:
    #         arr (list): a list of 32 byte hashes.
    #     """
    #     length = len(arr)
    #     self.write_var_int(length)
    #     for item in arr:
    #         ba = bytearray(binascii.unhexlify(item))
    #         ba.reverse()
    #         self.write_bytes(ba)

    def write_fixed8(self, value, unsigned=False):
        """
        Write a Fixed8 value to the stream.

        Args:
            value (neo.Fixed8):
            unsigned: (Not used)

        Returns:
            int: the number of bytes written
        """
        #        if unsigned:
        #            return self.write_u_int64(int(value.value))
        return self.write_int64(value.value)



class QiList(list):
    """
    QiList
    """
    @property
    def length(self):
        """
        返回序列的个数
        :return:
        """
        return len(self)

    @property
    def first(self):
        """
        序列的第一个元素
        :return:
        """
        if len(self) > 0:
            return self[0]
        else:
            return None

    @property
    def last(self):
        """
        序列的最后一个元素
        :return:
        """
        if len(self) > 0:
            return self[-1]
        else:
            return None

    @last.setter
    def last(self, value):
        if len(self) > 0:
            self[-1] = value

    @property
    def last_index(self):
        """
        返回当前序列的最后一个index
        :return:
        """
        return self.length - 1

    def ago(self, n):
        """
        返回当前序列从最后一个往前回溯指定个数的一个元素
        example:
        list = ['a', 'b', 'c', 'd']
        ago(0) = 'd'
        ago(1) = 'c'
        :param n:
        :return:
        """
        if n <= 0:
            return self.last
        if n >= (self.length - 1):
            return self.first
        return self[self.length - 1 - n]

    def ago_sub(self, n, m):
        """
        返回区间，包含头尾
        example:
        list = ['a', 'b', 'c', 'd']
        ago(0， 1) = ['c', 'd']
        ago(0, 2) = ['b', 'c', 'd']
        :param n:
        :param m:
        :return:
        """
        qi_list = QiList()
        num = max(n, m)
        num2 = min(n, m)
        num = max(0, num)
        num = min(self.length - 1, num)
        num2 = max(0, num2)
        num2 = min(self.length - 1, num2)
        for i in range(self.length - 1 - num, self.length - num2):
            qi_list.append(self[i])
        return qi_list

#
# data = QiList()
# data.append('a')
# data.append('b')
# data.append('c')
# data.append('d')
# data.append('e')
# data.append('f')
# data.append('g')
# print(data.first)
# print(data.last)
# print(data.last_index)
# print(data.ago(1))
# print(data.ago_sub(0, 1))



import enum
from enum import Enum, unique


@unique
class EnumExchange(enum.IntEnum):
    上证所 = 0
    深交所 = 1
    中金所 = 2
    上期所 = 3
    大商所 = 4
    郑商所 = 5
    新交所 = 6
    未知 = 7
    火币 = 8
    欧易 = 9  # OKEX
    币安 = 10




import datetime

from SapphireQuant.Core.CSharpUtils import CSharpUtils
from SapphireQuant.Core.Enum.EnumBarType import EnumBarType
from SapphireQuant.Core.Enum.EnumMarket import EnumMarket
from SapphireQuant.DataProcessing.DateTimeOffsetMap import DateTimeOffsetMap


class FileHeader:
    """
    头文件
    """
    FileVersionSize = 8
    FileHeaderSize = 64
    FileHeaderTotalSize = 64 + 32 * 20 + 32 * 20

    def __init__(self):
        self._isDirty = False
        self._file_version = '1.0'
        self._market = EnumMarket.期货
        self._bar_type = EnumBarType.minute
        self._begin_time = datetime.datetime(2999, 1, 1)
        self._end_time = datetime.datetime(2999, 1, 1)
        self._begin_trading_day = datetime.datetime(2999, 1, 1)
        self._end_trading_day = datetime.datetime(2999, 1, 1)
        self._bar_count = 0
        self._period = 1
        self._trading_day_indices = DateTimeOffsetMap(self)
        self._natural_day_indices = DateTimeOffsetMap(self)

    @property
    def file_version(self):
        """
        文件版本
        :return:
        """
        return self._file_version

    @file_version.setter
    def file_version(self, value):
        self._file_version = value

    @property
    def market(self):
        """
        市场
        :return:
        """
        return self._market

    @market.setter
    def market(self, value):
        self._market = value

    @property
    def bar_type(self):
        """
        分钟线类型
        :return:
        """
        return self._bar_type

    @bar_type.setter
    def bar_type(self, value):
        self._bar_type = value

    @property
    def begin_time(self):
        """
        开始时间
        :return:
        """
        return self._begin_time

    @begin_time.setter
    def begin_time(self, value):
        self._begin_time = value

    @property
    def end_time(self):
        """
        截止时间
        :return:
        """
        return self._end_time

    @end_time.setter
    def end_time(self, value):
        self._end_time = value

    @property
    def begin_trading_day(self):
        """
        开始交易日
        :return:
        """
        return self._begin_trading_day

    @begin_trading_day.setter
    def begin_trading_day(self, value):
        self._begin_trading_day = value

    @property
    def end_trading_day(self):
        """
        结束交易日
        :return:
        """
        return self._end_trading_day

    @end_trading_day.setter
    def end_trading_day(self, value):
        self._end_trading_day = value

    @property
    def bar_count(self):
        """

        :return:
        """
        return self._bar_count

    @bar_count.setter
    def bar_count(self, value):
        self._bar_count = value

    @property
    def period(self):
        """

        :return:
        """
        return self._period

    @period.setter
    def period(self, value):
        self._period = value

    @property
    def trading_day_indices(self):
        """

        :return:
        """
        return self._trading_day_indices

    @trading_day_indices.setter
    def trading_day_indices(self, value):
        self._trading_day_indices = value

    @property
    def natural_day_indices(self):
        """

        :return:
        """
        return self._natural_day_indices

    @natural_day_indices.setter
    def natural_day_indices(self, value):
        self._natural_day_indices = value

    def mark_as_dirty(self, value):
        """

        :param value:
        """
        self._isDirty = value

    def read(self, reader):
        """
        读取头
        :param reader:
        """
        reader.stream.seek(0, 0)
        self._file_version = reader.read_string()
        reader.stream.seek(self.FileVersionSize, 0)
        self._market = EnumMarket(reader.read_int32())
        self._bar_type = EnumBarType(reader.read_int32())
        self._period = reader.read_int32()
        self._begin_time = CSharpUtils.convert_c_sharp_ticks_to_py_date_time(reader.read_int64())
        self._end_time = CSharpUtils.convert_c_sharp_ticks_to_py_date_time(reader.read_int64())
        self._begin_trading_day = CSharpUtils.convert_c_sharp_ticks_to_py_date_time(reader.read_int64())
        self._end_trading_day = CSharpUtils.convert_c_sharp_ticks_to_py_date_time(reader.read_int64())
        self._bar_count = reader.read_int32()
        reader.stream.seek(self.FileHeaderSize, 0)
        self._trading_day_indices.read(reader)
        self._natural_day_indices.read(reader)
        self.mark_as_dirty(False)

    def write(self, writer):
        if self._isDirty:
            writer.stream.seek(0, 0)
            writer.write_var_string(self.file_version)
            writer.stream.seek(self.FileVersionSize, 0)
            writer.write_int32(self.market.value)
            writer.write_int32(self.bar_type.value)
            writer.write_int32(self.period)
            writer.write_int64(CSharpUtils.get_c_sharp_ticks(self.begin_time))
            writer.write_int64(CSharpUtils.get_c_sharp_ticks(self.end_time))
            writer.write_int64(CSharpUtils.get_c_sharp_ticks(self.begin_trading_day))
            writer.write_int64(CSharpUtils.get_c_sharp_ticks(self.end_trading_day))
            writer.write_int32(self.bar_count)
            writer.stream.seek(self.FileHeaderSize, 0)
            self._trading_day_indices.write(writer)
            self._natural_day_indices.write(writer)
            self.mark_as_dirty(False)
